/**********************************************************************************/
/*                                                                                */
/* File: main.cpp                                                                 */
/* Author: bkenwright@xbdev.net                                                   */
/* URL: www.xbdev.net                                                             */
/* Date: 08-09-2004                                                               */
/* Desciption: Simple mandelbrot fractal texture demo                             */
/*                                                                                */
/* Updated: 08-09-2004 by bkenwright                                              */
/*                                                                                */
/* Copyright (c) www.xbdev.net                                                    */
/*                                                                                */
/**********************************************************************************/
/*
   We show how easy it is to work with mandelbrot fractal textures!  And of course 
   how to create them.
   Mandelbrot fractal textures arn't used all that much in graphics...they create
   a sort of exciting random spherical effect...lots of noise etc.  One of those
   effects which you have to use your imagination for...such as maybe have it 
   altering in the background for a menu...have it for generating a volcano surface
   in a 3D world in a dynamic landscape...lots of things you can use it for if you
   put your mind to it.  And the best thing of all is its a very simple things to 
   learn and understand...and of course tweek and improve.
   
   All the mandelbrot code is contained in mandelbrot.cpp/.h

*/
/**********************************************************************************/

#define STRICT // Preprocessor variable, STRICT, which enforces the strictest 
               // possible type checking by the compiler 

#include <windows.h>
#include <stdio.h>

#pragma comment(lib, "D3d8.lib")
#pragma comment(lib, "D3dx8.lib")
#include <d3dx8.h>


#include "mandelbrot.h"      // CMandelbrot class
#include "meshobject.h"      // CMeshObject class


////////////////////////////////////////////////////////////////////////////////////
//                           Memory Leak Test Code                                //
////////////////////////////////////////////////////////////////////////////////////
//  Visual Studio debug, only tests for memory leaks with alloc(..) etc, the      //
//  standard C memory allocation functions.  It doens't give you much information //
//  when you use 'new' operator'  So we add these few extra lines to detect       //
//  any memory leaks.                                                             //
////////////////////////////////////////////////////////////////////////////////////
/*
//Standard method for detecting mem leaks with alloc(..)
#define _CRTDBG_MAP_ALLOC
#include <crtdbg.h>
_CrtDumpMemoryLeaks();
*/
////////////////////////////////////////////////////////////////////////////////////
#include <crtdbg.h>

static char THIS_FILE[] = __FILE__;

void* operator new(size_t nSize, const char * lpszFileName, int nLine)
{
    return ::operator new(nSize, 1, lpszFileName, nLine);
}
#define DEBUG_NEW new(THIS_FILE, __LINE__)

#define MALLOC_DBG(x) _malloc_dbg(x, 1, THIS_FILE, __LINE__);
#define malloc(x) MALLOC_DBG(x)

////////////////////////////////////////////////////////////////////////////////////
//                                                                                \\
// External Files                                                                 \\
//                                                                                \\
//\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\

#define FILE_X_OBJECT	"obj.x"

////////////////////////////////////////////////////////////////////////////////////
//                                                                                \\
// Global variables                                                               \\
//                                                                                \\
//\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\

IDirect3DDevice8	* g_pd3dDevice;
IDirect3D8			* g_pD3D;


////////////////////////////////////////////////////////////////////////////////////

// All the mandelbrot code is contained in CMandelbrot, and generates our mandelbrot 
// texture for us!  By putting it in this class its easier to follow it through 
// and see whats happening.
CMandelbrot g_mandelbrot;

// We've added a little .x loader class so we can also wrap our texture around
// something to see what its like
CMeshObject g_meshobject;



////////////////////////////////////////////////////////////////////////////////////
//                                                                                \\
// Name: Render(..)                                                               \\
// Date Created: 08-09-2004                                                       \\
// This little function is called over and over again, and does the rendering of  \\
// our simple 3D world.                                                           \\
//                                                                                \\
////////////////////////////////////////////////////////////////////////////////////
void Render()
{
    // Well if we need to do any drawing or rendering triangles we can do it in here!
	if(!g_pd3dDevice)return;

	
	////////////////////////////////////////////////////////////////////////////////
	//  Simple Camera                                                             //
	////////////////////////////////////////////////////////////////////////////////
	// We need to put our camera (aka our Eye) somewhere so we can look at our object
	// [1] D3DTS_VIEW
	D3DXMATRIX v;   
    g_pd3dDevice->SetTransform(D3DTS_VIEW, 
		                       D3DXMatrixLookAtLH(&v, 
							                      &D3DXVECTOR3(0,2,-35),    // eye
                                                  &D3DXVECTOR3(0,0,0),     // at
                                                  &D3DXVECTOR3(0,1,0)));    // up  
    // [2] D3DTS_PROJECTION
    D3DXMATRIX p;
    g_pd3dDevice->SetTransform( D3DTS_PROJECTION,  
		                        D3DXMatrixPerspectiveFovLH( &p,  D3DX_PI/4,  
								                                 1.0f,
                                                                 1.0f,
                                                                 500.0f));

	////////////////////////////////////////////////////////////////////////////////
	// Use the default world position of 0,0,0 and rotate our shape               //
	////////////////////////////////////////////////////////////////////////////////

	//// Just to rotate our shape!  See it rotate.
	static float angle = 0.0f;
    angle += 0.003f;
	//////~~~*~~~~ Create a matrix
	D3DXMATRIX mr;
	////~~~*~~~~ Do somthing to our empty matrix.
	D3DXMatrixRotationX(&mr, angle ); // angle in radians...eg. 1 degree = PI/180

	g_pd3dDevice->SetTransform( D3DTS_WORLD, &mr );
	


	// Clear the back buffer to a blue color
	g_pd3dDevice->Clear( 0, NULL, D3DCLEAR_TARGET|D3DCLEAR_ZBUFFER|D3DCLEAR_STENCIL, D3DCOLOR_XRGB(0,0,255), 1.0f, 0 );
	g_pd3dDevice->BeginScene();


	////////////////////////////////////////////////////////////////////////////////
	// Get our mandelbrot fractal texture so we can use it                        //
	////////////////////////////////////////////////////////////////////////////////
	IDirect3DTexture8 * pTexture;
	pTexture = g_mandelbrot.GetTexture(NULL);

	////////////////////////////////////////////////////////////////////////////////
	// Set our directx render states to how we want them                          //
	////////////////////////////////////////////////////////////////////////////////
	g_pd3dDevice->SetRenderState( D3DRS_ZENABLE, TRUE );
    g_pd3dDevice->SetRenderState( D3DRS_LIGHTING,  FALSE );
	g_pd3dDevice->SetRenderState( D3DRS_ALPHABLENDENABLE, FALSE);

	g_pd3dDevice->SetTextureStageState(0,D3DTSS_MAGFILTER,	D3DTEXF_LINEAR );
	g_pd3dDevice->SetTextureStageState(0,D3DTSS_MINFILTER,	D3DTEXF_LINEAR );

	////////////////////////////////////////////////////////////////////////////////
	// We need this! Make sure where rendering for a single texture...            //
	// You should really make sure your mesh is textured as well etc.             //
	////////////////////////////////////////////////////////////////////////////////
	g_pd3dDevice->SetTextureStageState(0,D3DTSS_COLOROP,	D3DTOP_SELECTARG1);
	g_pd3dDevice->SetTextureStageState(0,D3DTSS_COLORARG1,	D3DTA_TEXTURE);


	g_pd3dDevice->SetRenderState( D3DRS_CULLMODE, D3DCULL_CCW );

	// Use our mandelbrot fractal texture
	g_pd3dDevice->SetTexture( 0, pTexture );

	// Grab our loaded in mesh and render it ourselves with our own texture
	ID3DXMesh * pMesh = g_meshobject.GetMesh();

	// We use subset 0, which is usually the main material for the mesh.  You get
	// multiple subsets, e.g. 0, 1, 2 etc..if your mesh has lots of different
	// materials and textures.  But we know our mesh file is only a single mesh
	pMesh->DrawSubset(0);


	////////////////////////////////////////////////////////////////////////////////
	// Debug - Texture Information                                                //
	// Great! Amazing few lines, which if you think about it can be used for all  //
	// sorts of things...like rendering different camera views from different     //
	// positions etc.                                                             //
	////////////////////////////////////////////////////////////////////////////////
	#if 1
	{
		struct TLVERTEX
		{
			float x,y,z,rhw;
			float tu,tv;
		};
		#define	FVF_TLVERTEX (D3DFVF_XYZRHW | D3DFVF_TEX1)

		g_pd3dDevice->SetTextureStageState(0,D3DTSS_COLOROP,	D3DTOP_SELECTARG1);
		g_pd3dDevice->SetTextureStageState(0,D3DTSS_COLORARG1,	D3DTA_TEXTURE);
		float scale = 128.0f;
		TLVERTEX Vertex[4] = 
		{
			// x  y  z rhw tu tv
			{    0,    0,0, 1, 0, 0,},
			{scale,    0,0, 1, 1, 0,},
			{scale,scale,0, 1, 1, 1,},
			{    0,scale,0, 1, 0, 1,},
		};
		g_pd3dDevice->SetTexture( 0, pTexture );
		g_pd3dDevice->SetVertexShader( FVF_TLVERTEX );
		g_pd3dDevice->DrawPrimitiveUP( D3DPT_TRIANGLEFAN, 2, Vertex, sizeof( TLVERTEX ) );
	}
	#endif


	g_pd3dDevice->EndScene();
	// After rendering the scene we display it.
	g_pd3dDevice->Present( NULL, NULL, NULL, NULL );
}// End Render(..)


////////////////////////////////////////////////////////////////////////////////////
//                                                                                \\
//  Start() and End()                                                             \\
//  We call these two functions at the start and end of our directx setup.        \\
//  So we can allocate or release any memory or call any cleanup functions        \\
//  before releasing dx and exiting.                                              \\
//                                                                                \\
////////////////////////////////////////////////////////////////////////////////////

bool Start()
{
	HRESULT hr;
	//Create our mandelbrot fractal texture - its all done in this class
	hr = g_mandelbrot.Create( &g_pd3dDevice );
	if( hr!=S_OK ) return false;

	// Just for fun, we'll load in a mesh object, and wrap our texture around it
	// so we can see whta it looks like flat and on an object
	hr = g_meshobject.Create( &g_pd3dDevice, FILE_X_OBJECT, "");
	if( hr!=S_OK ) return false;

	return true;
}//End Start()


void End()
{
	// Always tidy up when you've finished.
	g_mandelbrot.Release();

	g_meshobject.Release();
}// End End()


////////////////////////////////////////////////////////////////////////////////////
//                                                                                \\
//  InitD3D is called once at the start of our program, and CleanUpDX is called   \\
//  once at the end of the program before closing.                                \\
//                                                                                \\
////////////////////////////////////////////////////////////////////////////////////

// Delete any allocated memory etc here.
void Cleanup()
{
	End();

	// Release DirectX3D resources.
	g_pd3dDevice->Release();
	g_pD3D->Release();
}//End Cleanup()



// Initializes Direct3DX at the start.
bool InitD3D(HWND hwnd)
{
    g_pD3D = Direct3DCreate8(D3D_SDK_VERSION);
	
	//Get the current display mode
    D3DDISPLAYMODE d3ddm;
    g_pD3D->GetAdapterDisplayMode(D3DADAPTER_DEFAULT, &d3ddm);
    //Create a structure to hold the settings for our device
    D3DPRESENT_PARAMETERS d3dpp; 
    ZeroMemory(&d3dpp, sizeof(d3dpp));

    //Fill the structure. 
    //We want our program to be windowed, and set the back buffer to a format
    //that matches our current display mode
    d3dpp.Windowed = TRUE;
    d3dpp.SwapEffect = D3DSWAPEFFECT_DISCARD;
    d3dpp.BackBufferFormat = d3ddm.Format;
    //For depth buffering (e.g.) the z-buffer
    d3dpp.BackBufferCount=1;
    //d3dpp.AutoDepthStencilFormat = D3DFMT_D16;
    d3dpp.EnableAutoDepthStencil = TRUE;

	// To get your windows width and height, remember it only takes this:
	int w = d3ddm.Width;
	int h = d3ddm.Height;

	// Lets get a list of what this graphics card can do for us :)
	D3DCAPS8 caps;
	g_pD3D->GetDeviceCaps(D3DADAPTER_DEFAULT, D3DDEVTYPE_HAL, &caps);

	// Check to see if our back buffer supports 32Bit z buffering - with
	// stencil buffer
	d3dpp.EnableAutoDepthStencil = TRUE;
	if( FAILED( g_pD3D->CheckDeviceFormat( caps.AdapterOrdinal, 
                                           caps.DeviceType, d3ddm.Format, 
                                           D3DUSAGE_DEPTHSTENCIL, 
                                           D3DRTYPE_SURFACE,
                                           D3DFMT_D24S8 ) ) )
	{
		d3dpp.AutoDepthStencilFormat = D3DFMT_D16;
	}
	else
	{
		d3dpp.AutoDepthStencilFormat = D3DFMT_D24S8;
	}

	// Ahhh...if we go full screen, its best to add this.
	//d3dpp.PresentationInterval = D3DPRESENT_INTERVAL_ONE;


	if (FAILED(g_pD3D->CreateDevice(D3DADAPTER_DEFAULT, D3DDEVTYPE_HAL,hwnd,D3DCREATE_HARDWARE_VERTEXPROCESSING,&d3dpp,&g_pd3dDevice))){
		if (FAILED(g_pD3D->CreateDevice(D3DADAPTER_DEFAULT, D3DDEVTYPE_HAL,hwnd,D3DCREATE_SOFTWARE_VERTEXPROCESSING,&d3dpp,&g_pd3dDevice))){
			if (FAILED(g_pD3D->CreateDevice(D3DADAPTER_DEFAULT, D3DDEVTYPE_REF,hwnd,D3DCREATE_SOFTWARE_VERTEXPROCESSING,&d3dpp,&g_pd3dDevice))){
					MessageBox(NULL,"Direct3D",  "Error Creating Device(..)" ,MB_OK | MB_ICONSTOP);
						return false;
			}
		}
	}// End CreateDevice(..) if block

	g_pd3dDevice->SetRenderState( D3DRS_ZENABLE, TRUE );


	// Make sure everything was loaded okay
	bool bGoodToGo = Start();
	if( bGoodToGo == false )
	{
		MessageBox(0, "Error setting up program 'Loading mesh object' Or 'Mandelbrot Fractal Image'",
					  "Settup Error in Start(..)", MB_OK | MB_ICONSTOP );
		return false;
	}


	return true;
}

////////////////////////////////////////////////////////////////////////////////////
//                                                                                \\
//  Our Windows Code....this is the windows code...I cramed it together so that   \\
//  it has its bare essentials...couldn't get it any smaller than this...so that  \\
//  we could concentrate at the heart of our coding....Octee and not the windows  \\
//  code.                                                                         \\
//                                                                                \\
////////////////////////////////////////////////////////////////////////////////////

LRESULT WINAPI MsgProc( HWND hWnd, UINT msg, WPARAM wParam, LPARAM lParam )
{
    switch( msg )
    {
        case WM_DESTROY:
			Cleanup();
            PostQuitMessage( 0 );
            return 0;
		case WM_RBUTTONDOWN:
			
			break;
    }

    return DefWindowProc( hWnd, msg, wParam, lParam );
}// End MsgProc(..)


// Program Entry Point
INT WINAPI WinMain( HINSTANCE hInst, HINSTANCE, LPSTR, INT )
{
    // Register the window class
	char szName[] = "www.xbdev.net";
    WNDCLASSEX wc = { sizeof(WNDCLASSEX), CS_CLASSDC, MsgProc, 0L, 0L, 
                      GetModuleHandle(NULL), NULL, NULL, NULL, NULL,
                      szName, NULL };
    RegisterClassEx( &wc );

    // Create the application's window
    HWND hWnd = CreateWindow( szName, szName, 
                              WS_OVERLAPPEDWINDOW, 100, 100, 600, 600,
                              GetDesktopWindow(), NULL, wc.hInstance, NULL );

    // Initialize Direct3DX
    if( InitD3D( hWnd ) == false ) return 0;
    
	// Show the window
    ShowWindow( hWnd, SW_SHOWDEFAULT );
    UpdateWindow( hWnd );

    // Enter the message loop
    MSG msg; 
    ZeroMemory( &msg, sizeof(msg) );
    while( msg.message!=WM_QUIT )
    {
		if( PeekMessage( &msg, NULL, 0U, 0U, PM_REMOVE ) )
        {
			TranslateMessage( &msg );
            DispatchMessage( &msg );
        }
        else
            Render();
    }

	_CrtDumpMemoryLeaks();

    // Clean up everything and exit the app
    UnregisterClass( szName, wc.hInstance );


    return 0;
}// End WinMain(..)








